@@ -1,11 +1,17 @@
 import asyncio
 import collections
+import warnings
 from typing import List  # noqa
-from typing import Awaitable, Callable, Optional, Tuple
+from typing import Awaitable, Callable, Generic, Optional, Tuple, TypeVar
 
-from .helpers import set_exception, set_result
+from .base_protocol import BaseProtocol
+from .helpers import BaseTimerContext, set_exception, set_result
 from .log import internal_logger
-from .typedefs import Byteish
+
+try:  # pragma: no cover
+    from typing import Deque  # noqa
+except ImportError:
+    from typing_extensions import Deque  # noqa
 
 
 __all__ = (
@@ -14,20 +20,22 @@ __all__ = (
 
 DEFAULT_LIMIT = 2 ** 16
 
+_T = TypeVar('_T')
+
 
 class EofStream(Exception):
     """eof stream indication."""
 
 
-class AsyncStreamIterator:
+class AsyncStreamIterator(Generic[_T]):
 
-    def __init__(self, read_func: Callable[[], Awaitable[bytes]]) -> None:
+    def __init__(self, read_func: Callable[[], Awaitable[_T]]) -> None:
         self.read_func = read_func
 
-    def __aiter__(self) -> 'AsyncStreamIterator':
+    def __aiter__(self) -> 'AsyncStreamIterator[_T]':
         return self
 
-    async def __anext__(self) -> bytes:
+    async def __anext__(self) -> _T:
         try:
             rv = await self.read_func()
         except EofStream:
@@ -54,17 +62,17 @@ class ChunkTupleAsyncStreamIterator:
 
 class AsyncStreamReaderMixin:
 
-    def __aiter__(self) -> AsyncStreamIterator:
+    def __aiter__(self) -> AsyncStreamIterator[bytes]:
         return AsyncStreamIterator(self.readline)  # type: ignore
 
-    def iter_chunked(self, n: int) -> AsyncStreamIterator:
+    def iter_chunked(self, n: int) -> AsyncStreamIterator[bytes]:
         """Returns an asynchronous iterator that yields chunks of size n.
 
         Python-3.5 available for Python 3.5+ only
         """
         return AsyncStreamIterator(lambda: self.read(n))  # type: ignore
 
-    def iter_any(self) -> AsyncStreamIterator:
+    def iter_any(self) -> AsyncStreamIterator[bytes]:
         """Returns an asynchronous iterator that yields all the available
         data as soon as it is received
 
@@ -98,8 +106,10 @@ class StreamReader(AsyncStreamReaderMixin):
 
     total_bytes = 0
 
-    def __init__(self, protocol,
-                 *, limit=DEFAULT_LIMIT, timer=None, loop=None):
+    def __init__(self, protocol: BaseProtocol,
+                 *, limit: int=DEFAULT_LIMIT,
+                 timer: Optional[BaseTimerContext]=None,
+                 loop: Optional[asyncio.AbstractEventLoop]=None) -> None:
         self._protocol = protocol
         self._low_water = limit
         self._high_water = limit * 2
@@ -108,17 +118,17 @@ class StreamReader(AsyncStreamReaderMixin):
         self._loop = loop
         self._size = 0
         self._cursor = 0
-        self._http_chunk_splits = None
-        self._buffer = collections.deque()
+        self._http_chunk_splits = None  # type: Optional[List[int]]
+        self._buffer = collections.deque()  # type: Deque[bytes]
         self._buffer_offset = 0
         self._eof = False
-        self._waiter = None
-        self._eof_waiter = None
-        self._exception = None
+        self._waiter = None  # type: Optional[asyncio.Future[bool]]
+        self._eof_waiter = None  # type: Optional[asyncio.Future[bool]]
+        self._exception = None  # type: Optional[BaseException]
         self._timer = timer
-        self._eof_callbacks = []
+        self._eof_callbacks = []  # type: List[Callable[[], None]]
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         info = [self.__class__.__name__]
         if self._size:
             info.append('%d bytes' % self._size)
@@ -198,9 +208,13 @@ class StreamReader(AsyncStreamReaderMixin):
         finally:
             self._eof_waiter = None
 
-    def unread_data(self, data: Byteish) -> None:
+    def unread_data(self, data: bytes) -> None:
         """ rollback reading some data from stream, inserting it to buffer head.
         """
+        warnings.warn("unread_data() is deprecated "
+                      "and will be removed in future releases (#3260)",
+                      DeprecationWarning,
+                      stacklevel=2)
         if not data:
             return
 
@@ -213,7 +227,7 @@ class StreamReader(AsyncStreamReaderMixin):
         self._eof_counter = 0
 
     # TODO: size is ignored, remove the param later
-    def feed_data(self, data: Byteish, size: int=0) -> None:
+    def feed_data(self, data: bytes, size: int=0) -> None:
         assert not self._eof, 'feed_data after feed_eof'
 
         if not data:
@@ -234,17 +248,40 @@ class StreamReader(AsyncStreamReaderMixin):
 
     def begin_http_chunk_receiving(self) -> None:
         if self._http_chunk_splits is None:
+            if self.total_bytes:
+                raise RuntimeError("Called begin_http_chunk_receiving when"
+                                   "some data was already fed")
             self._http_chunk_splits = []
 
     def end_http_chunk_receiving(self) -> None:
         if self._http_chunk_splits is None:
             raise RuntimeError("Called end_chunk_receiving without calling "
                                "begin_chunk_receiving first")
-        if not self._http_chunk_splits or \
-                self._http_chunk_splits[-1] != self.total_bytes:
-            self._http_chunk_splits.append(self.total_bytes)
 
-    async def _wait(self, func_name):
+        # self._http_chunk_splits contains logical byte offsets from start of
+        # the body transfer. Each offset is the offset of the end of a chunk.
+        # "Logical" means bytes, accessible for a user.
+        # If no chunks containig logical data were received, current position
+        # is difinitely zero.
+        pos = self._http_chunk_splits[-1] if self._http_chunk_splits else 0
+
+        if self.total_bytes == pos:
+            # We should not add empty chunks here. So we check for that.
+            # Note, when chunked + gzip is used, we can receive a chunk
+            # of compressed data, but that data may not be enough for gzip FSM
+            # to yield any uncompressed data. That's why current position may
+            # not change after receiving a chunk.
+            return
+
+        self._http_chunk_splits.append(self.total_bytes)
+
+        # wake up readchunk when end of http chunk received
+        waiter = self._waiter
+        if waiter is not None:
+            self._waiter = None
+            set_result(waiter, False)
+
+    async def _wait(self, func_name: str) -> None:
         # StreamReader uses a future to link the protocol feed_data() method
         # to a read coroutine. Running two read coroutines at the same time
         # would have an unexpected behaviour. It would not possible to know
@@ -356,13 +393,15 @@ class StreamReader(AsyncStreamReaderMixin):
                 return (b"", True)
             await self._wait('readchunk')
 
-        if not self._buffer:
+        if not self._buffer and not self._http_chunk_splits:
             # end of file
             return (b"", False)
         elif self._http_chunk_splits is not None:
             while self._http_chunk_splits:
                 pos = self._http_chunk_splits[0]
                 self._http_chunk_splits = self._http_chunk_splits[1:]
+                if pos == self._cursor:
+                    return (b"", True)
                 if pos > self._cursor:
                     return (self._read_nowait(pos-self._cursor), True)
             return (self._read_nowait(-1), False)
@@ -399,7 +438,7 @@ class StreamReader(AsyncStreamReaderMixin):
 
         return self._read_nowait(n)
 
-    def _read_nowait_chunk(self, n):
+    def _read_nowait_chunk(self, n: int) -> bytes:
         first_buffer = self._buffer[0]
         offset = self._buffer_offset
         if n != -1 and len(first_buffer) - offset > n:
@@ -421,7 +460,7 @@ class StreamReader(AsyncStreamReaderMixin):
             self._protocol.resume_reading()
         return data
 
-    def _read_nowait(self, n):
+    def _read_nowait(self, n: int) -> bytes:
         chunks = []
 
         while self._buffer:
@@ -443,7 +482,7 @@ class EmptyStreamReader(AsyncStreamReaderMixin):
     def set_exception(self, exc: BaseException) -> None:
         pass
 
-    def on_eof(self, callback: Callable[[], None]):
+    def on_eof(self, callback: Callable[[], None]) -> None:
         try:
             callback()
         except Exception:
@@ -461,7 +500,7 @@ class EmptyStreamReader(AsyncStreamReaderMixin):
     async def wait_eof(self) -> None:
         return
 
-    def feed_data(self, data: Byteish, n: int=0) -> None:
+    def feed_data(self, data: bytes, n: int=0) -> None:
         pass
 
     async def readline(self) -> bytes:
@@ -474,7 +513,7 @@ class EmptyStreamReader(AsyncStreamReaderMixin):
         return b''
 
     async def readchunk(self) -> Tuple[bytes, bool]:
-        return (b'', False)
+        return (b'', True)
 
     async def readexactly(self, n: int) -> bytes:
         raise asyncio.streams.IncompleteReadError(b'', n)
@@ -486,16 +525,16 @@ class EmptyStreamReader(AsyncStreamReaderMixin):
 EMPTY_PAYLOAD = EmptyStreamReader()
 
 
-class DataQueue:
+class DataQueue(Generic[_T]):
     """DataQueue is a general-purpose blocking queue with one reader."""
 
-    def __init__(self, *, loop=None):
+    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
         self._loop = loop
         self._eof = False
-        self._waiter = None
-        self._exception = None
+        self._waiter = None  # type: Optional[asyncio.Future[bool]]
+        self._exception = None  # type: Optional[BaseException]
         self._size = 0
-        self._buffer = collections.deque()
+        self._buffer = collections.deque()  # type: Deque[Tuple[_T, int]]
 
     def __len__(self) -> int:
         return len(self._buffer)
@@ -518,7 +557,7 @@ class DataQueue:
             set_exception(waiter, exc)
             self._waiter = None
 
-    def feed_data(self, data: Byteish, size: int=0) -> None:
+    def feed_data(self, data: _T, size: int=0) -> None:
         self._size += size
         self._buffer.append((data, size))
 
@@ -535,7 +574,7 @@ class DataQueue:
             self._waiter = None
             set_result(waiter, False)
 
-    async def read(self) -> bytes:
+    async def read(self) -> _T:
         if not self._buffer and not self._eof:
             assert not self._waiter
             self._waiter = self._loop.create_future()
@@ -555,29 +594,30 @@ class DataQueue:
             else:
                 raise EofStream
 
-    def __aiter__(self) -> AsyncStreamIterator:
+    def __aiter__(self) -> AsyncStreamIterator[_T]:
         return AsyncStreamIterator(self.read)
 
 
-class FlowControlDataQueue(DataQueue):
+class FlowControlDataQueue(DataQueue[_T]):
     """FlowControlDataQueue resumes and pauses an underlying stream.
 
     It is a destination for parsed data."""
 
-    def __init__(self, protocol, *, limit: int=DEFAULT_LIMIT, loop:
-                 Optional[asyncio.AbstractEventLoop]=None) -> None:
+    def __init__(self, protocol: BaseProtocol, *,
+                 limit: int=DEFAULT_LIMIT,
+                 loop: asyncio.AbstractEventLoop) -> None:
         super().__init__(loop=loop)
 
         self._protocol = protocol
         self._limit = limit * 2
 
-    def feed_data(self, data: Byteish, size: int=0) -> None:
+    def feed_data(self, data: _T, size: int=0) -> None:
         super().feed_data(data, size)
 
         if self._size > self._limit and not self._protocol._reading_paused:
             self._protocol.pause_reading()
 
-    async def read(self) -> bytes:
+    async def read(self) -> _T:
         try:
             return await super().read()
         finally:
